/*
 * Routines for dealing with crossover between MoE and MoM
 */
#include "lf_os_specific.h"

#include "libfma.h"
#include "lf_fabric.h"
#include "lf_fma_flags.h"
#include "lf_utility.h"

#include "fma.h"
#include "fma_fabric.h"
#include "fma_myrioe.h"
#include "fma_myri.h"
#include "fma_map.h"
#include "fma_standalone.h"

/*
 * local prototypes
 */

/*
 * Send a broadcast packet to each gateway in our Myrinet topo map,
 * if there is one.
 */
int
mome_broadcast_through_gateways(
  struct fma_nic_info *lcl_nip,
  int lcl_port,
  void *vpkt,
  int pkt_len,
  unsigned char *gw_mac_addr_location)
{
  struct mome_pkt_hdr *pkt;
  struct lf_fabric *fp;
  struct lf_host *hp;
  struct lf_nic *nicp;
  int num_sent;
  int rindex;
  int rlen;
  int rc;
  int h;
  int n;

  /* If there is no fabric struct, then nothing to do */
  fp = A.fabric;
  if (fp == NULL) return 0;

  pkt = vpkt;
  num_sent = 0;

  /* send a broadcast to each ethernet gateway */
  for (h=0; h<fp->num_hosts; ++h) {
    hp = fp->hosts[h];
    if (hp->fma_flags & FMA_FLAG_IS_GATEWAY) {
      for (n=0; n<hp->num_nics; ++n) {
	nicp = hp->nics[n];

	/* look through all my NICs for a route to this NIC */
	rindex = FMA_ROUTE_INDEX(lcl_nip->nic_id, lcl_port, 0);
	rlen = FMA_NIC(nicp)->route_lens[rindex];

	/* if no route from this NIC, move on to next gateway */
	if (rlen < 0) continue;

	/* copy gateway NIC into place if needed */
	if (gw_mac_addr_location != NULL) {
	  LF_MAC_COPY(gw_mac_addr_location, nicp->mac_addr);
	}

	/* send the packet */
	rc = mome_raw_send(lcl_nip, lcl_port,
		  FMA_NIC(nicp)->route_buf + FMA_ROUTE_OFFSET(rindex), rlen,
		  LF_BROADCAST_MAC, pkt, pkt_len, NULL, NULL);
	if (rc == -1) {
	  LF_ERROR(("Error sending MoEoM broadcast packet"));
	}
	++num_sent;
      }
    }
  }
  return num_sent;

 except:
  fma_perror_exit(1);
  return 0;
}

int
mome_raw_send(
  struct fma_nic_info *nip,
  int port,
  unsigned char *route,
  int rlen,
  lf_mac_addr_t dest_mac,
  void *vpkt,
  int pkt_len,
  void (*callback)(void *),
  void *arg)
{
  struct mome_pkt_hdr *pkt;
  uint32_t mac_lo;
  uint32_t mac_hi;
  int rc;

  pkt = vpkt;

  /* Fill MoE pkt header */
  pkt->h.m.proto_16 = htons(MYRI_TYPE_MX);

  mac_lo = lf_mac_low32(nip->myri_info.mac_addr);
  mac_hi = lf_mac_high16(nip->myri_info.mac_addr);
  pkt->h.m.src_mac_low_32 = htonl(mac_lo);
  pkt->h.m.src_mac_high_16 = htons(mac_hi);

  mac_lo = lf_mac_low16(dest_mac);
  mac_hi = lf_mac_high32(dest_mac);
  pkt->h.m.dest_mac_low_16 = htons(mac_lo);
  pkt->h.m.dest_mac_high_32 = htonl(mac_hi);

  pkt->h.m.sender_peer_index_16 = htons(0xFFFF);

  pkt->pkt_type = 1;		/* XXX */

  rc = fma_myri_raw_send(nip->nic_id, port, route, rlen,
                         pkt, pkt_len, callback, arg);

  return rc;
}

void
moe_check_mom_packet(
  struct fma_nic_info *nip,
  int port,
  struct mome_pkt_hdr *pkt,
  int length)
{
  int type;
  struct mome_pkt_hdr *mome_pkt;
  struct myri_ether_pkt *ether_pkt;

  type = ntohs(pkt->h.m.proto_16);

  /* converted MX packet? */
  if (type == MYRI_TYPE_MX) {
    if (pkt->pkt_type == 1) {	/* XXX */
      mome_handle_pkt(nip, port, LF_ZERO_MAC, (struct moe_pkt *)pkt, length);
    } else {
      fma_log("Got weird MX packet");
      lf_hex_dump(stdout, pkt, length);
    }

  /* EoM packet (== gatewayed EoM broadcast) ? */
  } else if (type == MYRI_TYPE_ETHER) {

    ether_pkt = (struct myri_ether_pkt *)pkt;
    mome_pkt = (struct mome_pkt_hdr *)
      ((char *)pkt + sizeof(struct myri_ether_pkt));

    if (mome_pkt->pkt_type == 1) {	/* XXX define! */
      mome_handle_pkt(nip, port, ether_pkt->sender_mac,
	  (struct moe_pkt *)mome_pkt, length - sizeof(struct myri_ether_pkt));
    } else {
      fma_log("Got non-RAW EoM packet");
      lf_hex_dump(stdout, pkt, length);
    }
  } else {
    fma_log("Not an EoM->MoM packet");
    lf_hex_dump(stdout, pkt, length);
  }
}
