#include "lf_os_specific.h"

#include "libfma.h"
#include "lf_scheduler.h"
#include "lf_channel.h"
#include "lf_fma_flags.h"
#include "lf_utility.h"

#include "fma.h"
#include "fma_fabric.h"
#include "fma_myrioe.h"
#include "fma_standalone.h"
#include "fma_myri.h"
#include "fma_map.h"
#include "fma_tunnel.h"
#include "fma_myri_packet.h"
#include "fma_map_fabric.h"

#define MOE_DEBUG 0

/*
 * Local prototypes
 */
static void fma_moe_handle_recv(struct fma_nic_info *nip,
  int port, struct moe_pkt *pkt, int pkt_len);
static void fma_moe_get_event(struct lf_channel *chp);
static void moe_broadcast_task(void *v);
static struct lf_nic *moe_find_peer(lf_mac_addr_t mac_addr);
static struct lf_nic *moe_find_or_add_peer(lf_mac_addr_t mac_addr,
  int num_ports, enum lf_firmware_type fw_type, int *new);
static void moe_got_bc_scout(struct fma_nic_info *nip, int port,
  lf_mac_addr_t gateway_mac, struct moe_bc_scout_pkt *pkt);
static void moe_got_bc_scout_reply(struct fma_nic_info *nip, int port,
  struct moe_bc_scout_reply_pkt *pkt);
static void moe_ack_scout_reply(struct fma_nic_info *nip, int port,
  struct lf_nic *peer_nic, int peer_port, int map_version);
static void moe_start_delay_expired(void *v);
static void moe_send_bc_scout_reply(struct lf_nic *peer_nicp, int peer_port,
  struct fma_nic_info *rx_nip, int rx_port, struct moe_bc_scout_pkt *bc_pkt);
static void moe_update_contact_info(struct lf_nic *peer_nicp, int rmt_port,
  struct fma_nic_info *nip, int lcl_port, lf_mac_addr_t gateway_mac);
static int moe_peer_raw_send(struct fma_nic_info *lcl_nip, int lcl_port,
  struct lf_nic *peer_nicp, int peer_port,
  void *vpkt, int pkt_len, void (*callback)(void *), void *arg);
static void moe_got_ack(struct fma_nic_info *nip, int port,
  struct moe_ack_pkt *pkt);
void moe_dump_peers(void);
static struct moe_scout_ack_record *moe_get_scout_ack_record( int nic_id,
  int port, lf_mac_addr_t scouter_mac, int scouter_port);
static void moe_verify_task(void *v);
static void moe_distribute_peer_table_to_peer(struct lf_nic *peer_nicp,
  int use_mom, struct moe_peer_list_pkt *pkt);
static void moe_mapper_watchdog_timeout(void *v);
static void moe_broadcast_verify_everywhere(void);
static int moe_broadcast_scout_everywhere(void);
static void moe_got_bc_verify(struct fma_nic_info *nip, int port,
  struct moe_bc_verify_pkt *pkt);
static void moe_request_remap(struct fma_nic_info *nip, int port,
  lf_mac_addr_t dest_mac, char *why);
static void moe_restart_mapper_watchdog(void);
static void moe_got_remap_request(struct fma_nic_info *nip, int port,
  struct moe_remap_request_pkt *pkt);
static void moe_clear_peer_table(void);
static void moe_dump_peer_table(void);

/*
 * Initialize myri-over-ethernet structs
 */
void
fma_init_myrioe_vars()
{
  struct fma_myrioe_info *moep;

  /* only allocate things once */
  if (A.myrioe != NULL) {
    return;
  }

  /* Allocate our info structure */
  LF_CALLOC(A.myrioe, struct fma_myrioe_info, 1);
  moep = A.myrioe;

  return;

 except:
  fma_perror_exit(1);
}

/*
 * NIC init code for Myrinet-over-Ethernet
 */
int
fma_init_nic_moe(
  int nic_id,
  int nic_handle)
{
  struct fma_nic_info *nip;
  struct fma_myrioe_info *moep;
  struct fma_myri *myp;
  struct lf_channel *chp;
  int delay;
  int rc;

  nip = NULL;
  chp = NULL;
  moep = A.myrioe;
  myp = A.myri;

  /* allocate and save a struct for this NIC */
  LF_CALLOC(nip, struct fma_nic_info, 1);
  myp->nic_info[myp->fma_num_nics] = nip;

  /* fill in fields */
  nip->nic_handle = nic_handle;
  nip->nic_id = nic_id;
  nip->nic_index = myp->fma_num_nics;
  nip->ni_net_type = MYRI_NT_ETHERNET;

  /* get Myrinet information for NIC */
  rc = myri_get_nic_info(nic_handle, &nip->myri_info);
  if (rc == -1) LF_ERROR(("Getting NIC info for nic_id"));

  /* print a nice message for each NIC found */
  fma_log("NIC %d: MoE %s s/n=%s %d ports, speed=%dG", nic_id,
      nip->myri_info.product_id,
      nip->myri_info.serial_no,
      nip->myri_info.num_ports,
      nip->myri_info.speed==MYRI_SPEED_2G?2:10);
  fma_log("       mac = " LF_MAC_FORMAT, LF_MAC_ARGS(nip->myri_info.mac_addr));

  /* save MAX mac address on this host */
  if (LF_MAC_CMP(nip->myri_info.mac_addr, A.myri->my_max_mac) > 0) {
    LF_MAC_COPY(A.myri->my_max_mac, nip->myri_info.mac_addr);
  }

  /* allocate space for counters */
  LF_CALLOC(nip->last_counters, struct myri_nic_counters,
            nip->myri_info.num_ports);

  /* See if we need to set the hostname for the NIC */
  fma_check_nic_hostname(nic_handle);

  /* allocate channel for this NIC */
  LF_CALLOC(chp, struct lf_channel, 1);
  nip->chp = chp;

  /* fill in the channel */
  chp->fd = myri_fd(nic_handle);
  chp->hangup_rtn = fma_nic_hangup;
  chp->context = nip;

  /* add this channel to poll queue */
  if (lf_add_channel(chp) != 0) {
    LF_FREE(chp);
    LF_ERROR(("Error adding channel"));
  }

  /* register callback for events */
  lf_channel_receive(chp, NULL, 0, fma_moe_get_event);

  /* Schedule a mapping start */
  if (moep->moe_start_delay_timer == NULL) {
    delay = lf_random() % MOE_MAP_START_DELAY;
    moep->moe_start_delay_timer = lf_schedule_event(moe_start_delay_expired,
						NULL, delay);
    if (moep->moe_start_delay_timer == NULL) {
      LF_ERROR(("Error starting start delay timer"));
    }

    ++myp->queries_in_progress;		/* XXX - pretend this is a query */
  }

  ++myp->fma_num_nics;	/* increase num_nics once everything has succeeded */
  ++myp->moe_num_nics;

  return 0;

 except:
  return -1;
}

static void
moe_start_delay_expired(
  void *v)
{
  /* clear the event pointer */
  A.myrioe->moe_start_delay_timer = NULL;

  if (MOE_DEBUG || A.debug) {
    fma_log("MoE start delay expired");
  }

  /* treat this as a query completion to fit into myri framework */
  fma_note_query_completion();
}

/*
 * Start the mapping process.
 * Clear all peer_nic_masks in peer table to say that no-one is accessible.
 * Update our map version and broadcast a query, collect replies.
 * Frequently re-broadcast query until no new replies are seen, making sure
 * we broadcast at least a minumum number of times.  Once all broadcasts are
 * complete, start sending out the maps.
 */
void
moe_start_mapping()
{
  struct fma_myrioe_info *moep;
  struct fma_myri *myp;
  struct fma_nic_info *nip;
  struct lf_nic *peer_nicp;
  int n;
  int p;

  moep = A.myrioe;
  myp = A.myri;

  /* stop verifying */
  LF_CLEAR_EVENT(moep->moe_verify_task);

  /* stop watchdog timer just in case */
  LF_CLEAR_EVENT(moep->moe_mapper_watchdog);

  fma_log("MoE mapping starting now.");
  moep->moe_mapping_in_progress = TRUE;
  moep->moe_map_version = A.myri->map_session_id;

  moep->moe_broadcast_count = 0;	/* no broadcasts yet */

  /* Clear out the peer table */
  moe_clear_peer_table();

  /* ensure that all of our MoE NICs are in the peer table */
  for (n=0; n<myp->fma_num_nics; ++n) {
    nip = myp->nic_info[n];
    if (nip->ni_net_type == MYRI_NT_ETHERNET) {
      peer_nicp = moe_find_or_add_peer(nip->myri_info.mac_addr, 
				  nip->myri_info.num_ports,
				  myri_firmware_type(), NULL);

      /* save the pointer to our nip struct */
      nip->nic_ptr = peer_nicp;
      FMA_NIC(peer_nicp)->local_nip = nip;
      FMA_NIC(peer_nicp)->fn_level = A.stand->my_level;

      /* set up routes to self */
      for (p=0; p<peer_nicp->num_ports; ++p) {
	moe_update_contact_info(peer_nicp, p, nip, p, LF_ZERO_MAC);
      }
    }
  }

  /* Go ahead and do broadcasts on all NICs */
  moep->moe_broadcast_task = lf_schedule_event(moe_broadcast_task, NULL, 0);
  if (moep->moe_broadcast_task == NULL) {
    LF_ERROR(("Error arming MoE broadcast timer"));
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Stop MoE mapping - usually called because we encountered 
 * a higher-level mapper
 */
void
moe_stop_mapping()
{
  struct fma_myrioe_info *moep;

  moep = A.myrioe;

  /* clear the broadcast task event if active */
  LF_CLEAR_EVENT(moep->moe_broadcast_task);

  moep->moe_mapping_in_progress = FALSE;	/* mapping not in progress */
}

/*
 * Time to read an event from the NIC in Myri-over-Ether mode
 */
static void
fma_moe_get_event(
  struct lf_channel *chp)
{
  struct fma_nic_info *nip;
  struct myri_event *mep;
  int rc;

  nip = chp->context;		/* get pointer to NIC info */

  /* get the next event for this NIC */
  rc = myri_next_event(nip->nic_handle, &mep, 0);
  if (rc != 0) LF_ERROR(("Error getting event from NIC"));

  switch (mep->type) {
    case MYRI_EVENT_RAW_RECV_COMPLETE:
      fma_moe_handle_recv(nip, mep->d.raw_recv.port,
	  (struct moe_pkt *)mep->d.raw_recv.rxbuf, mep->d.raw_recv.rx_len);
      break;

    case MYRI_EVENT_RAW_SEND_COMPLETE:
      fma_myri_send_complete(mep->d.raw_send.context);
      break;

    case MYRI_EVENT_ERROR:
      fma_myri_error(nip, mep->d.error.error);
      break;

    case MYRI_EVENT_NO_EVENT:
      LF_ERROR(("No event when there should have been one!?"));
      break;

    default:
      break;
  }
  myri_release_event(mep);
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Got a packet on MoE interface
 */
static void
fma_moe_handle_recv(
  struct fma_nic_info *nip,
  int port,
  struct moe_pkt *pkt, 
  int pkt_len)
{
  /* handle the packet, no gateway known */
  mome_handle_pkt(nip, port, LF_ZERO_MAC, pkt, pkt_len);
}

/*
 * Got an MoE-style packet on some interface
 */
void
mome_handle_pkt(
  struct fma_nic_info *nip,
  int port,
  lf_mac_addr_t gateway_mac,
  struct moe_pkt *pkt, 
  int pkt_len)
{
  enum fma_packet_subtype pkt_subtype;

  /* extract and switch on packet subtype */
  pkt_subtype = ntohs(pkt->moe_pkt_subtype_16);

  switch (pkt_subtype) {

  case FMA_SUBTYPE_MOE_BC_SCOUT:
    moe_got_bc_scout(nip, port, gateway_mac, (struct moe_bc_scout_pkt *)pkt);
    break;

  case FMA_SUBTYPE_MOE_BC_SCOUT_REPLY:
    moe_got_bc_scout_reply(nip, port, (struct moe_bc_scout_reply_pkt *)pkt);
    break;

  case FMA_SUBTYPE_MOE_ACK:
    moe_got_ack(nip, port, (struct moe_ack_pkt *)pkt);
    break;

  case FMA_SUBTYPE_MOE_BC_VERIFY:
    moe_got_bc_verify(nip, port, (struct moe_bc_verify_pkt *)pkt);
    break;

  case FMA_SUBTYPE_MOE_REMAP_REQUEST:
    moe_got_remap_request(nip, port, (struct moe_remap_request_pkt *)pkt);
    break;

  case FMA_SUBTYPE_FMA_TUNNEL_START:
    if (nip->ni_net_type == MYRI_NT_MYRINET) {
      fma_tunnel_got_start(nip, port,
	  (struct fma_myri_packet *) &pkt->moe_pkt_type_16,
	  fma_moeom_generic_raw_send);
    } else {
      fma_tunnel_got_start(nip, port,
	  (struct fma_myri_packet *) &pkt->moe_pkt_type_16,
	  fma_moe_generic_raw_send);
    }
    break;

  case FMA_SUBTYPE_FMA_TUNNEL_BODY:
    fma_tunnel_got_body((struct fma_myri_packet *) &pkt->moe_pkt_type_16);
    break;

  case FMA_SUBTYPE_FMA_TUNNEL_ACK:
    fma_tunnel_got_ack((struct fma_myri_packet *) &pkt->moe_pkt_type_16);
    break;

  default:
    fma_log("Unknown MoE pkt subtype %d", pkt_subtype);
    lf_hex_dump(stdout, pkt, pkt_len);
    break;
  }
}

/*
 * Got a request to remap fabric
 * Pass off a pointer to the fma_map_request embedded within
 */
static void
moe_got_remap_request(
  struct fma_nic_info *nip,
  int port,
  struct moe_remap_request_pkt *pkt)
{
  fma_peer_remap_request(&pkt->fma_remap_request, pkt->moe_requester_mac_addr);
}

/*
 * Got an ACK for a scout reply.
 * If this matches, clear the associated ack_pending flag
 */
static void
moe_got_ack(
  struct fma_nic_info *nip,
  int port,
  struct moe_ack_pkt *pkt)
{
  struct moe_scout_ack_record *sap;
  struct lf_nic *peer_nicp;
  int map_version;

  if (A.debug) {
    fma_log("Got BC scout ACK from " LF_MAC_FORMAT ", port %d",
	LF_MAC_ARGS(pkt->moe_scouter_mac_addr), pkt->moe_tx_port);
  }

  peer_nicp = moe_find_peer(pkt->moe_scouter_mac_addr);
  if (peer_nicp == NULL) {
    fma_log("ACK from unknown peer");
    return;
  }

  map_version = ntohl(pkt->moe_map_version_32);

  /* get pointer to scout ACK record */
  sap = moe_get_scout_ack_record(nip->nic_id, port,
      pkt->moe_scouter_mac_addr, pkt->moe_tx_port);

  if (sap->sa_map_version == map_version) {
    if (MOE_DEBUG || A.debug) {
      fma_log("Got matching scout reply ACK");
    }
    sap->sa_map_ack_pending = FALSE;
  } else {
    fma_log("Scout reply ACK map version %d != %d", map_version, 
	sap->sa_map_version);
  }
}

/*
 * Received a broadcast scout.
 * If sender is of higher-level than us and new mapper or new
 * map version, then reply.  Stop any mapping in progress.
 * If we are higher level, then start mapping if not already mapping.
 * (XXX could optimize to simple unicast reply - reply, add him to map,
 * and re-send map to all or even just him)
 */
static void
moe_got_bc_scout(
  struct fma_nic_info *nip,
  int port,
  lf_mac_addr_t gateway_mac,
  struct moe_bc_scout_pkt *pkt)
{
  struct fma_myrioe_info *moep;
  struct moe_scout_ack_record *sap;
  struct lf_nic *peer_nicp;
  struct fma_myri *myp;
  struct fma_map_info *mip;
  int map_version;
  int peer_port;
  int new;
  int level;

  map_version = ntohl(pkt->moe_map_version_32);
  level = pkt->moe_level_8;

  if (MOE_DEBUG || A.debug) {
    fma_log("nic %d Got BC scout from " LF_MAC_FORMAT", vers=%d", nip->nic_id,
	LF_MAC_ARGS(pkt->moe_scouter_mac_addr), map_version);
  }

  moep = A.myrioe;
  myp = A.myri;
  mip = A.map_info;

  /* make sure he is in our peer table */
  peer_nicp = moe_find_or_add_peer(pkt->moe_scouter_mac_addr, 
      pkt->moe_num_ports, pkt->moe_fw_type, &new);
  peer_port = pkt->moe_tx_port;
  FMA_NIC(peer_nicp)->fn_level = level;

  /* update "route" information for this peer */
  moe_update_contact_info(peer_nicp, peer_port,
      nip, port, gateway_mac);

  /* get pointer to scout ACK record */
  sap = moe_get_scout_ack_record(nip->nic_id, port,
      peer_nicp->mac_addr, peer_port);

  /* If this is new map version or ACK is pending still, reply */
  if (map_version != sap->sa_map_version || sap->sa_map_ack_pending) {

      moe_send_bc_scout_reply(peer_nicp, pkt->moe_tx_port, nip, port, pkt);
  }

  /* Hand off to the "got contact" engine */
  fma_standalone_got_contact(level, pkt->moe_scouter_max_mac_addr,
      nip, port, NULL, 0, pkt->moe_scout_type);

  /* If we are not the mapper, restart mapper watchdog */
  if (!A.stand->active) {
    moe_restart_mapper_watchdog();
  }
}

#if 0
  /* If new mapper, or new map version, or we have a pending ack, reply! */
  cmp = fma_compare_levels(level, pkt->moe_scouter_max_mac_addr,
			   mip->mi_mapper_level, mip->mi_mapper_mac_addr);

  level = pkt->moe_level_8;
  map_version = ntohl(pkt->moe_map_version_32);

  cmp = fma_compare_levels(level, pkt->moe_scouter_max_mac_addr,
      			   A.stand->my_level, A.myri->my_max_mac);

  /* Is this guy bigger than we are? */
  if (cmp > 0) {

    if (A.debug) {
      fma_log("Got BC scout from bigger peer, " LF_MAC_FORMAT ", level = %d",
	  LF_MAC_ARGS(pkt->moe_scouter_max_mac_addr), level);
    }

    if (moep->moe_mapping_in_progress) {
      moe_stop_mapping();		/* stop mapping */
    }

    /* stop verifying */
    LF_CLEAR_EVENT(moep->moe_verify_task);

  /* We are bigger than this node -
   * if not mapping
   *   start mapping
   * if mapping
   *   ignore
   */
  } else if (cmp < 0) {

    if (myp->fma_num_nics == 0
	|| fma_compare_levels(A.map_level, myp->my_max_mac,
	  mip->mi_mapper_level, mip->mi_mapper_mac_addr) >= 0) {

      fma_log("Got scout from smaller peer");
      fma_standalone_map_is_invalid("Got MoE scout from smaller peer");
    } else {
      fma_log("Ignoring smaller scout because i am not MoM mapper");
    }

  /* Got scout from self - reply! */
  } else {

    /* don't do anything if it came back on the same port */
    if (LF_MAC_CMP(pkt->moe_scouter_mac_addr, nip->myri_info.mac_addr) == 0
	&& pkt->moe_tx_port == port) {
      fma_log("NIC %d port %d got own scout!", nip->nic_id, port);
      return;
    }

    moe_send_bc_scout_reply(peer_nicp, pkt->moe_tx_port, nip, port, pkt);
  }
}
#endif

static void
moe_mapper_watchdog_timeout(
  void *v)
{
  struct fma_myrioe_info *moep;
  struct fma_standalone_data *sdp;

  moep = A.myrioe;
  sdp = A.stand;

  moep->moe_mapper_watchdog = NULL;		/* timer popped */

  fma_log("Mapper watchdog timer popped!");

  /* Become the active mapper in this case */
  sdp->active = TRUE;

  /* make me be the mapper */
  fma_set_mapper(A.myri->my_max_mac, sdp->my_level);

  /* This will report it and start a remap */
  fma_standalone_map_is_invalid("Timeout waiting for mapper scout");
}

/*
 * Send a broadcast scout reply packet
 */
static void
moe_send_bc_scout_reply(
  struct lf_nic *peer_nicp,
  int peer_port,
  struct fma_nic_info *rx_nip,
  int rx_port,
  struct moe_bc_scout_pkt *bc_pkt)
{
  struct moe_bc_scout_reply_pkt pkt;
  struct moe_scout_ack_record *sap;
  int map_version;
  int rc;

  map_version = ntohl(bc_pkt->moe_map_version_32);

  /* get pointer to scout ACK record */
  sap = moe_get_scout_ack_record(rx_nip->nic_id, rx_port,
      peer_nicp->mac_addr, peer_port);

  if (sap->sa_map_version == map_version
      && !sap->sa_map_ack_pending) {
    if (A.debug > 1) {
      fma_log("Reply already sent and ACKed for map version %d", map_version);
    }
    return;
  }

  if (MOE_DEBUG || A.debug) {
    fma_log("Sending scout reply to " LF_MAC_FORMAT ", gw=" LF_MAC_FORMAT,
	LF_MAC_ARGS(bc_pkt->moe_scouter_mac_addr),
	LF_MAC_ARGS(bc_pkt->moe_gateway_mac_addr));
  }

  pkt.moe_pkt_subtype_16 = htons(FMA_SUBTYPE_MOE_BC_SCOUT_REPLY);
  pkt.moe_map_version_32 = bc_pkt->moe_map_version_32;
  LF_MAC_COPY(pkt.moe_scouter_mac_addr, bc_pkt->moe_scouter_mac_addr);
  LF_MAC_COPY(pkt.moe_scoutee_mac_addr, rx_nip->myri_info.mac_addr);
  LF_MAC_COPY(pkt.moe_scoutee_max_mac_addr, A.myri->my_max_mac);
  LF_MAC_COPY(pkt.moe_gateway_mac_addr, bc_pkt->moe_gateway_mac_addr);
  pkt.moe_fw_type = myri_firmware_type();
  pkt.moe_num_ports = rx_nip->myri_info.num_ports;
  pkt.moe_rx_port = rx_port;
  pkt.moe_scoutee_level_8 = A.stand->my_level;

  /* Expect an ACK for this reply */
  sap->sa_map_version = map_version;
  sap->sa_map_ack_pending = TRUE;

  /* send reply pkt back on same NIC/port */
  rc = moe_peer_raw_send(rx_nip, rx_port, peer_nicp, peer_port,
      			 &pkt, sizeof(pkt), NULL, NULL);
  if (rc != 0) {
    LF_ERROR(("Error sending scout reply pkt"));
  }
  return;

 except:
  fma_perror_exit(1);
}


static void
moe_got_bc_scout_reply(
  struct fma_nic_info *nip,
  int port,
  struct moe_bc_scout_reply_pkt *pkt)
{
  struct lf_nic *peer_nicp;
  struct fma_myrioe_info *moep;
  int map_version;
  int new;

  if (MOE_DEBUG || A.debug) {
    fma_log("NIC %d got scout reply from " LF_MAC_FORMAT, nip->nic_id,
	LF_MAC_ARGS(pkt->moe_scoutee_mac_addr));
  }

  moep = A.myrioe;
  moep->moe_map_reply_flag = TRUE;	/* note that we got a reply */

  /* see if this NIC knows this peer, add if not */
  peer_nicp = moe_find_or_add_peer(pkt->moe_scoutee_mac_addr,
      pkt->moe_num_ports, pkt->moe_fw_type, &new);
  FMA_NIC(peer_nicp)->fn_level = pkt->moe_scoutee_level_8;

  if (new) {
    if (MOE_DEBUG || A.debug) {
      fma_log("Got scout reply from new peer " LF_MAC_FORMAT,
	  LF_MAC_ARGS(pkt->moe_scoutee_mac_addr));
    }
  }

  moe_update_contact_info(peer_nicp, pkt->moe_rx_port, nip, port, 
      pkt->moe_gateway_mac_addr);

  /* Make sure this reply is for current map version */
  map_version = ntohl(pkt->moe_map_version_32);
  if (map_version != moep->moe_map_version) {

    fma_log("Stale map version in scout reply - %d should be %d",
	map_version, moep->moe_map_version);

  } else {

    /* If we are actively mapping, just ACK the reply */
    if (moep->moe_mapping_in_progress) {

      /* ACK the scout reply */
      moe_ack_scout_reply(nip, port, peer_nicp, pkt->moe_rx_port, map_version);

    } else {
      if (A.debug) {
	fma_log("late MoE scout reply from "LF_MAC_FORMAT, 
	    LF_MAC_ARGS(pkt->moe_scoutee_mac_addr));
      }
    }
  }

  /* Note that we got contact from this guy */
  fma_standalone_got_contact(pkt->moe_scoutee_level_8,
      pkt->moe_scoutee_max_mac_addr,
      nip, port, NULL, 0, FMA_NST_VERIFY);
}

static void
moe_ack_scout_reply(
  struct fma_nic_info *nip,
  int port,
  struct lf_nic *peer_nicp,
  int peer_port,
  int map_version)
{
  struct moe_ack_pkt pkt;
  int rc;

  if (MOE_DEBUG || A.debug) {
    fma_log("ACKing map scout reply for %u to " LF_MAC_FORMAT,
	map_version, LF_MAC_ARGS(peer_nicp->mac_addr));
  }

  pkt.moe_pkt_subtype_16 = htons(FMA_SUBTYPE_MOE_ACK);
  pkt.moe_map_version_32 = htonl(map_version);
  LF_MAC_COPY(pkt.moe_scouter_mac_addr, nip->myri_info.mac_addr);
  pkt.moe_tx_port = port;

  rc = moe_peer_raw_send(nip, port, peer_nicp, peer_port, &pkt, sizeof(pkt),
      NULL, NULL);
  if (rc == -1) {
    LF_ERROR(("Error sending scout reply ACK"));
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * MoE mapping all done!
 */
static void
moe_mapping_done()
{
  struct fma_myrioe_info *moep;
  struct fma_map_progress *mpp;

  moep = A.myrioe;
  mpp = A.myri->map_progress;

  moep->moe_mapping_in_progress = FALSE;

  fma_log("MoE mapping took %d seconds",
      (int)(time(NULL) - mpp->mf_start_time));

  /* start verify going if there is anything to verify */
  if (moep->moe_num_peers > 0) {
    moep->moe_verify_task = lf_schedule_event(moe_verify_task, NULL,
	MOE_VERIFY_BROADCAST_INTERVAL);
    if (moep->moe_verify_task == NULL) {
      LF_ERROR(("Error arming verify timer"));
    }
  }

  fma_mf_mapping_complete();
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Do a mapping broadcast
 */
static void
moe_broadcast_task(
  void *v)
{
  struct fma_myrioe_info *moep;
  int num_sent;

  moep = A.myrioe;

  moep->moe_broadcast_task = NULL;		/* timer popped */

  /* If we got no replies and enough broadcasts have been sent, 
   * mapping is done!
   */
  if (!moep->moe_map_reply_flag
      && moep->moe_broadcast_count > MOE_MAP_BROADCAST_COUNT) {
    moe_mapping_done();

  /* not done, we will broadcast (at least) once more */
  } else {
    ++moep->moe_broadcast_count;
    moep->moe_map_reply_flag = FALSE;

    /* broadcast our MAC on all NICs */
    num_sent = moe_broadcast_scout_everywhere();

    /* only keep going if we actually sent anything */
    if (num_sent > 0) {
      /* re-arm timer */
      moep->moe_broadcast_task = lf_schedule_event(moe_broadcast_task,
	  NULL, MOE_MAP_BROADCAST_INTERVAL);
      if (moep->moe_broadcast_task == NULL) {
	LF_ERROR(("Error re-arming MoE broadcast timer"));
      }
    } else {
      fma_log("No MoE targets, MoE mapping done");
      moe_mapping_done();
    }
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Build a broadcast scout packet and send it everywhere
 */
static int
moe_broadcast_scout_everywhere()
{
  struct moe_bc_scout_pkt pkt;
  struct fma_myrioe_info *moep;
  struct fma_nic_info *nip;
  struct fma_myri *myp;
  int num_sent;
  int port;
  int n;
  int rc;

  moep = A.myrioe;
  myp = A.myri;

  num_sent = 0;

  /* build packet body */
  pkt.moe_pkt_subtype_16 = htons(FMA_SUBTYPE_MOE_BC_SCOUT);
  LF_MAC_COPY(pkt.moe_scouter_max_mac_addr, A.myri->my_max_mac);
  pkt.moe_level_8 = A.stand->my_level;
  pkt.moe_map_version_32 = htonl(moep->moe_map_version);
  pkt.moe_fw_type = myri_firmware_type();
  pkt.moe_scout_type = A.map_info->first_mapping ? FMA_NST_MAP : FMA_NST_REMAP;

  /* Do one broadcast to each NIC */
  for (n=0; n<myp->fma_num_nics; ++n) {
    nip = myp->nic_info[n];

    LF_MAC_COPY(pkt.moe_scouter_mac_addr, nip->myri_info.mac_addr);
    pkt.moe_num_ports = nip->myri_info.num_ports;

    for (port=0; port<nip->myri_info.num_ports; ++port) {

      if (MOE_DEBUG || A.debug) {
	fma_log("Sending MoE BC scout through NIC %d, port %d",
	    nip->nic_id, port);
      }

      pkt.moe_tx_port = port;

      /* If this is an MoE NIC, then directly broadcast */
      if (nip->ni_net_type == MYRI_NT_ETHERNET) {

	/* Not through gateway, so put ZERO in gw mac addr */
	LF_MAC_COPY(pkt.moe_gateway_mac_addr, LF_ZERO_MAC);

	/* Send the packet */
	rc = moe_raw_send(nip, port, LF_BROADCAST_MAC, &pkt, sizeof(pkt),
	    NULL, NULL);
	if (rc != 0) {
	  LF_ERROR(("Error sending MyrioE broadcast scout"));
	}
	++num_sent;

      /* MoM NIC - broadcast through all gateways in fabric */
      } else {
	num_sent += mome_broadcast_through_gateways(nip, port,
	    &pkt, sizeof(pkt), pkt.moe_gateway_mac_addr);
      }
    }
  }
  return num_sent;

 except:
  fma_perror_exit(1);
  return 0;
}


/*
 * Send a packet to a peer, either directly or through a gateway
 */
static int
moe_peer_raw_send(
  struct fma_nic_info *lcl_nip,
  int lcl_port,
  struct lf_nic *peer_nicp,
  int peer_port,
  void *vpkt,
  int pkt_len,
  void (*callback)(void *),
  void *arg)
{
  int rindex;
  int gw_rindex;
  unsigned char *gw_mac;
  struct lf_nic *gw_nicp;

  /* find route index */
  rindex = FMA_ROUTE_INDEX(lcl_nip->nic_id, lcl_port, peer_port);

  /* If 0-length route, then direct send */
  if (FMA_NIC(peer_nicp)->route_lens[rindex] == 0) {
    return moe_raw_send(lcl_nip, lcl_port, peer_nicp->mac_addr,
	vpkt, pkt_len, callback, arg);
  } else if (FMA_NIC(peer_nicp)->route_lens[rindex] == sizeof(lf_mac_addr_t)) {

    /* get MAC addr of gateway */
    gw_mac = FMA_NIC(peer_nicp)->route_buf + FMA_ROUTE_OFFSET(rindex);

    /* get NIC struct for gateway */
    gw_nicp = lf_find_nic_by_mac(A.fabric, gw_mac);

    /* If we cannot find the gateway, just drop pkt.  The gateway will
     * show up later
     */
    if (gw_nicp == NULL) {
      if (MOE_DEBUG || A.debug) {
	fma_log("Dropping send to unknown gateway " LF_MAC_FORMAT, 
	    LF_MAC_ARGS(gw_mac));
      }
      return 0;
    }

    /* Now find route to gateway */
    /* XXX - assumes 0 port # for gateway */
    gw_rindex = FMA_ROUTE_INDEX(lcl_nip->nic_id, lcl_port, 0);

    return mome_raw_send(lcl_nip, lcl_port, 
	FMA_NIC(gw_nicp)->route_buf + FMA_ROUTE_OFFSET(rindex),
	FMA_NIC(gw_nicp)->route_lens[rindex],
	peer_nicp->mac_addr, vpkt, pkt_len, callback, arg);

  /* Bad route length... */
  } else {
    fma_log("Bad route len for " LF_MAC_FORMAT" is %d",
	LF_MAC_ARGS(peer_nicp->mac_addr),
	FMA_NIC(peer_nicp)->route_lens[rindex]);
    return 0;
  }
}

/*
 * Perform a raw send on Myri-over-ethernet
 */
int
moe_raw_send(
  struct fma_nic_info *nip,
  int port,
  lf_mac_addr_t dest_mac_addr,
  void *vpkt,
  int pkt_len,
  void (*callback)(void *),
  void *arg)
{
  struct moe_pkt *pkt;
  int rc;

  pkt = vpkt;

  /* build packet header */
  memset(&pkt->hdr, 0, sizeof(struct mome_pkt_hdr));
  LF_MAC_COPY(pkt->hdr.h.e.dest_mac, dest_mac_addr);
  LF_MAC_COPY(pkt->hdr.h.e.src_mac, nip->myri_info.mac_addr);
  pkt->hdr.h.e.proto_16 = htons(ETHER_TYPE_MX);
  pkt->hdr.h.e.sender_peer_index_16 = htons(0xFFFF);
  pkt->hdr.pkt_type = 1;

  /* Send the packet */
  rc = fma_myri_raw_send(nip->nic_handle, port, NULL, 0, pkt, pkt_len,
      callback, arg);

  return rc;
}

static void
moe_clear_peer_table()
{
  struct fma_myrioe_info *moep;
  struct lf_nic *peer_nicp;
  int p;

  moep = A.myrioe;

  /* Look through all the peer NICS for this MAC */
  for (p=0; p<moep->moe_num_peers; ++p) {

    /* Clear any nic_info backpointer */
    peer_nicp = moep->moe_peers[p];
    if (FMA_NIC(peer_nicp)->local_nip != NULL) {
      FMA_NIC(peer_nicp)->local_nip->nic_ptr = NULL;
    }
    lf_free_nic(peer_nicp);

    moep->moe_peers[p] = NULL;
  }

  moep->moe_num_peers = 0;
}

static struct lf_nic *
moe_find_peer(
  lf_mac_addr_t mac_addr)
{
  struct fma_myrioe_info *moep;
  struct lf_nic *peer_nicp;
  int p;

  moep = A.myrioe;

  /* Look through all the peer NICS for this MAC */
  for (p=0; p<moep->moe_num_peers; ++p) {
    peer_nicp = moep->moe_peers[p];
    if (LF_MAC_CMP(peer_nicp->mac_addr, mac_addr) == 0) {
      return peer_nicp;
    }
  }

  return NULL;
}

/*
 * Look through the list of peers to see if we can find one
 * that this NIC has seen before.  Add the entry as having been seen
 * by this NIC if not.
 */
static struct lf_nic *
moe_find_or_add_peer(
  lf_mac_addr_t mac_addr,
  int num_ports,
  enum lf_firmware_type fw_type,
  int *new)
{
  struct fma_myrioe_info *moep;
  struct lf_nic *peer_nicp;

  moep = A.myrioe;

  /* Look through all the peer NICS for this MAC */
  peer_nicp = moe_find_peer(mac_addr);

  /* If not found, add the entry */
  if (peer_nicp == NULL) {

    /* If table is not big enough, grow it now */
    if (moep->moe_peers_size <= moep->moe_num_peers) {
      moep->moe_peers_size += 32;
      moep->moe_peers = realloc(moep->moe_peers, 
	  (moep->moe_peers_size * sizeof(struct lf_nic *)));
      if (moep->moe_peers == NULL) {
	LF_ERROR(("Error growing peer table"));
      }
    }

    /* Grab the next entry */
    peer_nicp = fma_create_nic(num_ports, mac_addr, fw_type);
    moep->moe_peers[moep->moe_num_peers] = peer_nicp;
    ++moep->moe_num_peers;

    if (new != NULL) {
      *new = TRUE;		/* it is new! */
    }

  /* If we found it, not new */
  } else {
    if (new != NULL) {
      *new = FALSE;		/* peer already existed for this NIC */
    }
  }

  return peer_nicp;

 except:
  fma_perror_exit(1);
  return NULL;
}

/*
 * Update the route information for this NIC.  If we are going through 
 * a gateway, then the route is the MAC address of the gateway. We will delay
 * figuring out the route to the gateway until we actually send a packet
 * since the route may change over time.
 *
 * If we are directly connected on same ethernet, then it is a 0-length route.
 */
static void
moe_update_contact_info(
  struct lf_nic *peer_nicp,
  int rmt_port,
  struct fma_nic_info *nip,
  int lcl_port,
  lf_mac_addr_t gateway_mac)
{
  int rindex;
  unsigned char *route;

  if (MOE_DEBUG) {
    fma_log("Adding route from NIC %d to " LF_MAC_FORMAT,
	nip->nic_id, LF_MAC_ARGS(peer_nicp->mac_addr));
  }

  /* find route index */
  rindex = FMA_ROUTE_INDEX(nip->nic_id, lcl_port, rmt_port);

  /* If there is a gateway, that is the route */
  if (LF_MAC_CMP(gateway_mac, LF_ZERO_MAC) != 0) {
    route = FMA_NIC(peer_nicp)->route_buf + FMA_ROUTE_OFFSET(rindex);
    FMA_NIC(peer_nicp)->route_lens[rindex] = sizeof(lf_mac_addr_t);
    LF_MAC_COPY(route, gateway_mac);

  /* no gateway - route is NULL */
  } else {
    FMA_NIC(peer_nicp)->route_lens[rindex] = 0;
  }
}

/*
 * Get the record for keeping track of ACKs to this scouter
 */
static struct moe_scout_ack_record *
moe_get_scout_ack_record(
  int nic_id,
  int port,
  lf_mac_addr_t scouter_mac,
  int scouter_port)
{
  struct moe_scout_ack_record *sap;
  struct fma_myrioe_info *moep;

  moep = A.myrioe;

  sap = moep->scout_acks;
  while (sap != NULL) {
    if (sap->sa_rx_nic_id == nic_id
	&& sap->sa_rx_port == port
	&& sap->sa_scouter_port == scouter_port
	&& LF_MAC_CMP(sap->sa_scouter_mac, scouter_mac) == 0) {
      return sap;
    }
    sap = sap->next;
  }

  LF_CALLOC(sap, struct moe_scout_ack_record, 1);
  sap->sa_rx_nic_id = nic_id;
  sap->sa_rx_port = port;
  sap->sa_scouter_port = scouter_port;
  LF_MAC_COPY(sap->sa_scouter_mac, scouter_mac);

  /* put ad head of list */
  sap->next = moep->scout_acks;
  moep->scout_acks = sap;

  return sap;

 except:
  fma_perror_exit(1);
  return NULL;
}

/*
 * verify task -
 * the mapper needs to send a periodic broadcast
 * everywhere to ensure that folks know we are alive and to catch any
 * new nodes that come online.
 */
static void
moe_verify_task(
  void *v)
{
  struct fma_myrioe_info *moep;

  moep = A.myrioe;

  moep->moe_verify_task = NULL;		/* timer popped */

  /* If we are no longer the mapper, stop verifying */
  if (LF_MAC_CMP(A.myri->my_max_mac, A.map_info->mi_mapper_mac_addr) != 0) {
    fma_log("No longer mapper, stopping MoE verification");
    return;
  }

  /* send a broadcast with current map version */
  moe_broadcast_verify_everywhere();

  /* re-arm verify timer */
  moep->moe_verify_task = lf_schedule_event(moe_verify_task, NULL,
      MOE_VERIFY_BROADCAST_INTERVAL);
  if (moep->moe_verify_task == NULL) {
    LF_ERROR(("Error re-arming verify timer"));
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Generic raw send for MoE
 * We will need to copy the packet into a buffer with room for
 * the prepended MoE header.
 */
int
fma_moe_generic_raw_send(
  struct fma_generic_send_info *gsip,
  void *p,
  int len,
  void (*callback)(void *),
  void *context)
{
  unsigned char pkt_buf[8192];

  /* copy packet in just after where the header will be */
  memcpy(pkt_buf + sizeof(struct mome_pkt_hdr), p, len);

  return moe_raw_send(gsip->gsi_nip, gsip->gsi_port, gsip->gsi_dest_mac_addr, 
      pkt_buf, len + sizeof(struct mome_pkt_hdr), callback, context);
}
/*
 * Generic raw send for MoEoM
 * We will need to copy the packet into a buffer with room for
 * the prepended MoE header.
 */
int
fma_moeom_generic_raw_send(
  struct fma_generic_send_info *gsip,
  void *p,
  int len,
  void (*callback)(void *),
  void *context)
{
  unsigned char pkt_buf[8192];

  /* copy packet in just after where the header will be */
  memcpy(pkt_buf + sizeof(struct mome_pkt_hdr), p, len);

  return mome_raw_send(gsip->gsi_nip, gsip->gsi_port,
      gsip->gsi_route, gsip->gsi_route_len,
      gsip->gsi_dest_mac_addr,
      pkt_buf, len + sizeof(struct mome_pkt_hdr), callback, context);
}

/*
 * Build the distributable peer table
 */
void
moe_build_peer_table()
{
  struct fma_myrioe_info *moep;
  struct lf_nic *peer_nicp;
  struct moe_peer_entry *pep;
  struct lf_fabric *fp;
  int tbl_len;
  int p;

  moep = A.myrioe;
  fp = A.fabric;
  
  /* do nothing if no MoE peers */
  if (moep->moe_num_peers == 0) return;

  tbl_len = moep->moe_num_peers;
  if (fp != NULL) {
    tbl_len += FMA_FABRIC(fp)->af_num_nics;
  }

  LF_FREE(moep->moe_peer_table);
  LF_CALLOC(moep->moe_peer_table, struct moe_peer_entry, tbl_len);

  pep = moep->moe_peer_table;
  for (p=0; p<moep->moe_num_peers; ++p) {
    peer_nicp = moep->moe_peers[p];

    /* don't add fabric NICs */
    if (lf_find_nic_by_mac(fp, peer_nicp->mac_addr) != NULL) {
      continue;
    }

    /* garbage for directly connected NICs, but we don't care */
    LF_MAC_COPY(pep->pe_gateway_mac,
	FMA_NIC(peer_nicp)->route_buf + FMA_ROUTE_OFFSET(0));
    LF_MAC_COPY(pep->pe_peer_mac, peer_nicp->mac_addr);
    pep->pe_level_8 = FMA_NIC(peer_nicp)->fn_level;
    pep->pe_num_ports = peer_nicp->num_ports;
    pep->pe_fw_type = FMA_NIC(peer_nicp)->fw_type;
    ++pep;
  }

  if (fp != NULL) {
    for (p=0; p<FMA_FABRIC(fp)->af_num_nics; ++p) {
      peer_nicp = FMA_FABRIC(fp)->af_nics[p];

      /* don't bother with nodes that cannot to MoE */
      if (!(peer_nicp->host->fma_flags & FMA_FLAG_CAN_DO_MOE)) {
	continue;
       }

      /* garbage for directly connected NICs, but we don't care */
      LF_MAC_COPY(pep->pe_gateway_mac, LF_ZERO_MAC);
      LF_MAC_COPY(pep->pe_peer_mac, peer_nicp->mac_addr);
      pep->pe_level_8 = FMA_MOM_LEVEL_BONUS;
      pep->pe_num_ports = peer_nicp->num_ports;
      pep->pe_fw_type = FMA_NIC(peer_nicp)->fw_type;
      ++pep;
    }
  }

  moep->moe_peer_table_len = pep - moep->moe_peer_table;

  moe_dump_peer_table();
  return;

 except:
  fma_perror_exit(1);
}

static void
moe_dist_peer_done(
  void *vpeer)
{
  struct lf_nic *peer_nicp;

  peer_nicp = vpeer;
  fma_log("MoE map dist done to " LF_MAC_FORMAT,
      LF_MAC_ARGS(peer_nicp->mac_addr));
}

static void
moe_dist_peer_failed(
  void *vpeer)
{
  struct lf_nic *peer_nicp;

  peer_nicp = vpeer;
  fma_log("Map dist FAILED to " LF_MAC_FORMAT,
      LF_MAC_ARGS(peer_nicp->mac_addr));
}

/*
 * Distribute the peer table
 */
void
moe_distribute_peer_table()
{
  struct fma_myrioe_info *moep;
  struct moe_peer_list_pkt pkt;
  struct lf_nic *peer_nicp;
  struct lf_fabric *fp;
  int p;

  moep = A.myrioe;
  fp = A.fabric;

  /* Version of map that is loaded */
  moep->moe_map_version = A.myri->map_session_id;
  
  /* do nothing if no MoE peers */
  if (moep->moe_num_peers == 0) return;

  pkt.hdr.type_16 = htons(FMA_PACKET_TYPE);
  pkt.hdr.subtype_16 = htons(FMA_SUBTYPE_MOE_PEER_LIST);
  pkt.moe_num_peers_32 = htonl(moep->moe_peer_table_len);
  pkt.moe_map_version_32 = htonl(moep->moe_map_version);

  for (p=0; p<moep->moe_num_peers; ++p) {

    peer_nicp = moep->moe_peers[p];

    /* Skip local NICs */
    if (FMA_NIC(peer_nicp)->local_nip != NULL) {
      continue;
    }

    /* Skip if in fabric also */
    if (lf_find_nic_by_mac(fp, peer_nicp->mac_addr) != NULL) {
      continue;
    }

    moe_distribute_peer_table_to_peer(peer_nicp, FALSE, &pkt);
  }
  
  if (fp != NULL) {
    for (p=0; p<FMA_FABRIC(fp)->af_num_nics; ++p) {

      peer_nicp = FMA_FABRIC(fp)->af_nics[p];

      /* Skip local NICs */
      if (FMA_NIC(peer_nicp)->local_nip != NULL) {
	continue;
      }

      /* don't bother with nodes that cannot to MoE */
      if (!(peer_nicp->host->fma_flags & FMA_FLAG_CAN_DO_MOE)) {
	continue;
      }

      moe_distribute_peer_table_to_peer(peer_nicp, TRUE, &pkt);
    }
  }
}

static void
moe_distribute_peer_table_to_peer(
  struct lf_nic *peer_nicp,
  int use_mom,
  struct moe_peer_list_pkt *pkt)
{
  struct fma_myri *myp;
  struct fma_nic_info *nip;
  unsigned char *route;
  struct fma_myrioe_info *moep;
  struct lf_fabric *fp;
  struct lf_nic *gw_nicp;
  int lcl_port;
  int rmt_port;
  int rindex;
  int n;
  int msg_len;
  int rlen;

  moep = A.myrioe;
  myp = A.myri;
  fp = A.fabric;
  msg_len = moep->moe_peer_table_len * sizeof(struct moe_peer_entry);

  /* Scan through NIC ports until we find one through which we
   * can distribute this peer table.
   */
  for (n=0; n<myp->fma_num_nics; ++n) {
    nip = A.myri->nic_info[n];
    for (lcl_port=0; lcl_port<nip->myri_info.num_ports; ++lcl_port) {

      for (rmt_port=0; rmt_port<peer_nicp->num_ports; ++rmt_port) {
	rindex = FMA_ROUTE_INDEX(n, lcl_port, rmt_port);

	/* If no route here, keep looking */
	rlen = FMA_NIC(peer_nicp)->route_lens[rindex];
	if (rlen == -1) continue;

	route = FMA_NIC(peer_nicp)->route_buf + FMA_ROUTE_OFFSET(rindex);

	if (nip->ni_net_type == MYRI_NT_MYRINET) {

	  /* Use MoM to reach this peer? */
	  if (use_mom) {

	    if (MOE_DEBUG) fma_log("  using MoM NIC %d, port %d", n, lcl_port);

	    fma_tunnel_send(pkt, sizeof(*pkt),
		moep->moe_peer_table, msg_len,
		route, rlen,
		nip, lcl_port,
		moe_dist_peer_done, moe_dist_peer_failed,
		peer_nicp);

	    return;

	  /* No, require a gateway route, then */
	  } else {

	    if (rlen != 6) {
	      fma_log("Route from NIC %d, p%d to " LF_MAC_FORMAT " is not 6!",
		  n, lcl_port, LF_MAC_ARGS(peer_nicp->mac_addr));
	      continue;
	    }

	    /* If we can't find the gateway, try next NIC port */
	    gw_nicp = lf_find_nic_by_mac(fp, route);
	    if (gw_nicp == NULL) {
	      continue;
	    }

	    /* XXX - only checking for route to port 0 of gateway */
	    rindex = FMA_ROUTE_INDEX(n, lcl_port, 0);
	    route = FMA_NIC(gw_nicp)->route_buf + FMA_ROUTE_OFFSET(rindex);
	    rlen = FMA_NIC(gw_nicp)->route_lens[rindex];

	    if (rlen == -1) {
	      continue;
	    }

	    if (MOE_DEBUG) {
	      fma_log("  using MoE NIC %d, port %d thru "LF_MAC_FORMAT,
		  n, lcl_port, LF_MAC_ARGS(gw_nicp->mac_addr));
	    }

	    fma_tunnel_send_generic(pkt, sizeof(*pkt),
		moep->moe_peer_table, msg_len,
		peer_nicp->mac_addr,
		route, rlen,
		nip, lcl_port, fma_moeom_generic_raw_send,
		moe_dist_peer_done, moe_dist_peer_failed,
		peer_nicp);
	    return;
	  }

	/* Sending from MoE NIC - just use the peer MAC addr */
	} else {

	  if (MOE_DEBUG) fma_log("  using MoM NIC %d, port %d", n, lcl_port);

	  fma_tunnel_send_generic(pkt, sizeof(*pkt),
	      moep->moe_peer_table, msg_len,
	      peer_nicp->mac_addr,
	      NULL, 0,
	      nip, lcl_port, fma_moe_generic_raw_send,
	      moe_dist_peer_done, moe_dist_peer_failed,
	      peer_nicp);
	  return;
	}
      }
    }
  }

  if (MOE_DEBUG) {
    fma_log("Could not find route to distribute peer table to "LF_MAC_FORMAT,
	LF_MAC_ARGS(peer_nicp->mac_addr));
  }
}

/*
 * Got a peer list - load up the peer table and routes
 */
void
moe_got_peer_list(
  struct moe_peer_list_pkt *pkt,
  lf_mac_addr_t sender_mac)
{
  int num_entries;
  struct moe_peer_entry *pep;
  struct fma_myrioe_info *moep;
  struct fma_nic_info *nip;
  struct lf_nic *peer_nicp;
  struct lf_fabric *fp;
  struct fma_myri *myp;
  int rindex;
  int p;
  int n;

  fp = A.fabric;
  myp = A.myri;
  moep = A.myrioe;

  /* save map version */
  moep->moe_map_version = ntohl(pkt->moe_map_version_32);

  /* clear existing peer table */
  moe_clear_peer_table();

  num_entries = ntohl(pkt->moe_num_peers_32);
  fma_log("got peer list from " LF_MAC_FORMAT ", version=%u, num_entries=%d",
      LF_MAC_ARGS(sender_mac), moep->moe_map_version,  num_entries);

  moep->moe_last_rank = 0;
  
  /* get access to start of entry array */
  pep = (struct moe_peer_entry *)&pkt[1];
  for (p=0; p<num_entries; ++p) {

    if (fma_compare_levels(A.stand->my_level, myp->my_max_mac,
	  		   pep->pe_level_8, pep->pe_peer_mac) < 0) {
      ++moep->moe_last_rank;
    }

    peer_nicp = moe_find_or_add_peer(pep->pe_peer_mac,
			pep->pe_num_ports, pep->pe_fw_type, NULL);
    FMA_NIC(peer_nicp)->fn_level = pep->pe_level_8;

    /* set up route info for each of our NICs */
    for (n=0; n<myp->fma_num_nics; ++n) {
      nip = myp->nic_info[n];
      rindex = FMA_ROUTE_INDEX(n, 0, 0);

      if (nip->ni_net_type == MYRI_NT_MYRINET) {

	/* only bother with peers not in fabric */
	if (lf_find_nic_by_mac(fp, peer_nicp->mac_addr) == NULL) {
	  FMA_NIC(peer_nicp)->route_lens[rindex] = 6;
	  LF_MAC_COPY(FMA_NIC(peer_nicp)->route_buf + FMA_ROUTE_OFFSET(rindex),
	      pep->pe_gateway_mac);
	}

      /* MoE NIC - 0-length route */
      } else {
	FMA_NIC(peer_nicp)->route_lens[rindex] = 0;
      }
    }
    ++pep;
  }

  fma_set_all_routes();

  LF_CLEAR_EVENT(A.stand->map_wait_timer);
}

/*
 * Build a broadcast verify packet and send it everywhere
 */
static void
moe_broadcast_verify_everywhere()
{
  struct moe_bc_verify_pkt pkt;
  struct fma_myrioe_info *moep;
  struct fma_nic_info *nip;
  struct fma_myri *myp;
  int port;
  int n;
  int rc;

  moep = A.myrioe;
  myp = A.myri;

  /* build packet body */
  pkt.moe_pkt_subtype_16 = htons(FMA_SUBTYPE_MOE_BC_VERIFY);
  LF_MAC_COPY(pkt.moe_verifier_max_mac_addr, A.myri->my_max_mac);
  pkt.moe_map_version_32 = htonl(moep->moe_map_version);
  pkt.moe_verifier_level_8 = A.stand->my_level;

  /* Do one broadcast to each NIC */
  for (n=0; n<myp->fma_num_nics; ++n) {
    nip = myp->nic_info[n];
    LF_MAC_COPY(pkt.moe_verifier_mac_addr, nip->myri_info.mac_addr);

    for (port=0; port<nip->myri_info.num_ports; ++port) {

      if (MOE_DEBUG || A.debug) {
	fma_log("MoE BC verify through NIC %d, port %d", nip->nic_id, port);
      }

      /* If this is an MoE NIC, then directly broadcast */
      if (nip->ni_net_type == MYRI_NT_ETHERNET) {

	/* Send the packet */
	rc = moe_raw_send(nip, port, LF_BROADCAST_MAC, &pkt, sizeof(pkt),
	    NULL, NULL);
	if (rc != 0) {
	  LF_ERROR(("Error sending MyrioE broadcast scout"));
	}

      /* MoM NIC - broadcast through all gateways in fabric */
      } else {
	(void) mome_broadcast_through_gateways(nip, port,
					       &pkt, sizeof(pkt), NULL);
      }
    }
  }
  return;

 except:
  fma_perror_exit(1);
}

static void
moe_restart_mapper_watchdog()
{
  struct fma_myrioe_info *moep;
  struct fma_myri *myp;

  moep = A.myrioe;
  myp = A.myri;

  /* clear running timer */
  LF_CLEAR_EVENT(moep->moe_mapper_watchdog);

  /* restart watchdog timer only if we have some MoE NICs 
   * non-MoE NICs will not (and do not need to) see mapper verify broadcasts
   */
  if (myp->moe_num_nics > 0) {
    moep->moe_mapper_watchdog = lf_schedule_event(
	moe_mapper_watchdog_timeout, NULL,
	MOE_MAPPER_WATCHDOG_TIMEOUT(moep->moe_last_rank));
    if (moep->moe_mapper_watchdog == NULL) {
      LF_ERROR(("Error starting mapper watchdog timer"));
    }
  }
  return;

 except:
  fma_perror_exit(1);
}

static void
moe_got_bc_verify(
  struct fma_nic_info *nip,
  int port,
  struct moe_bc_verify_pkt *pkt)
{
  struct fma_myrioe_info *moep;
  struct fma_myri *myp;
  struct fma_map_info *mip;
  int map_version;

  map_version = ntohl(pkt->moe_map_version_32);

  if (MOE_DEBUG || A.debug) {
    fma_log("nic %d Got BC verify from " LF_MAC_FORMAT, nip->nic_id,
	LF_MAC_ARGS(pkt->moe_verifier_mac_addr));
    fma_log("  max_mac= "LF_MAC_FORMAT", version=%u",
	LF_MAC_ARGS(pkt->moe_verifier_mac_addr), map_version);

  }

  moep = A.myrioe;
  myp = A.myri;
  mip = A.map_info;

  /* If this is from our mapper, clear the watchdog timer */
  if (LF_MAC_CMP(pkt->moe_verifier_max_mac_addr,
	mip->mi_mapper_mac_addr) == 0) {

    if (map_version == moep->moe_map_version) {
      moe_restart_mapper_watchdog();

    } else {
      fma_log("Got verify from mapper with map vers %d != %d",
	  map_version, moep->moe_map_version);

      /* Complain if map version is wrong */
      /* (only applies to MoE NICs - MoM NICs have their own way */
      if (nip->ni_net_type == MYRI_NT_ETHERNET) {
	moe_request_remap(nip, port, pkt->moe_verifier_mac_addr,
	    "Verify with bad map version");
      }
    }

  /* Got verify from non-mapper... XXX */
  } else {
    if (MOE_DEBUG || A.debug) {
      fma_log("Verify packet is from non-mapper ", LF_MAC_FORMAT,
	  LF_MAC_ARGS(pkt->moe_verifier_mac_addr));
    }
    fma_standalone_got_contact(pkt->moe_verifier_level_8,
	pkt->moe_verifier_max_mac_addr,
	nip, port, NULL, 0, FMA_NST_VERIFY);

    /* If this did not cause us to become active, request remap */
    /* (only applies to MoE NICs - MoM NICs have their own way */
    if (nip->ni_net_type == MYRI_NT_ETHERNET) {
      moe_request_remap(nip, port, pkt->moe_verifier_mac_addr,
	  "Verify from non-mapper");
    }
  }
}

/*
 * Request a remap from another node
 */
static void
moe_request_remap(
  struct fma_nic_info *nip,
  int port,
  lf_mac_addr_t dest_mac,
  char *why)
{
  struct moe_remap_request_pkt pkt;
  struct fma_myrioe_info *moep;
  int rc;

  moep = A.myrioe;

  pkt.moe_pkt_subtype_16 = htons(FMA_SUBTYPE_MOE_REMAP_REQUEST);
  LF_MAC_COPY(pkt.moe_requester_mac_addr, nip->myri_info.mac_addr);

  /* fill in fma_map_request portion */
  pkt.fma_remap_request.map_version_32 = htonl(moep->moe_map_version);
  strcpy(pkt.fma_remap_request.why, why);

  rc = moe_raw_send(nip, port, dest_mac, &pkt, sizeof(pkt), NULL, NULL);
  if (rc == -1) {
    LF_ERROR(("Error sending remap request"));
  }
  return;

 except:
  fma_perror_exit(1);
}

void
moe_dump_peers()
{
  struct fma_myrioe_info *moep;
  struct fma_myri *myp;
  struct fma_nic_info *lcl_nip;
  struct lf_nic *peer_nicp;
  int rindex;
  int peer_port;
  int lcl_port;
  int rlen;
  int n;
  int p;

  moep = A.myrioe;
  myp = A.myri;

  fma_log("%d peers", moep->moe_num_peers);
  for (p=0; p<moep->moe_num_peers; ++p) {
    peer_nicp = moep->moe_peers[p];
    fma_log(LF_MAC_FORMAT, LF_MAC_ARGS(peer_nicp->mac_addr));
    for (peer_port=0; peer_port<peer_nicp->num_ports; ++peer_port) {
      fma_log("  port %d", peer_port);
      for (n=0; n<myp->fma_num_nics; ++n) {
	lcl_nip = myp->nic_info[n];
	for (lcl_port=0; lcl_port < lcl_nip->myri_info.num_ports; ++lcl_port) {
	  rindex = FMA_ROUTE_INDEX(n, lcl_port, peer_port);
	  rlen = FMA_NIC(peer_nicp)->route_lens[rindex];
	  if (rlen == 0) {
	    fma_log("    NIC %d, p%d: direct", n, lcl_port);
	  } else if (rlen == -1) {
	    fma_log("    NIC %d, p%d: no path", n, lcl_port);
	  } else {
	    fma_log("    NIC %d, p%d: " LF_MAC_FORMAT, n, lcl_port,
		LF_MAC_ARGS(FMA_NIC(peer_nicp)->route_buf
		            + FMA_ROUTE_OFFSET(rindex)));
	  }
	}
      }
    }
  }
}

static void
moe_dump_peer_table()
{
  struct fma_myrioe_info *moep;
  struct moe_peer_entry *pep;
  int p;

  moep = A.myrioe;

  fma_log("Peer table (map_version=%u) :", moep->moe_map_version);
  for (p=0; p<moep->moe_peer_table_len; ++p) {
    pep = &moep->moe_peer_table[p];
    fma_log("  %d " LF_MAC_FORMAT " gw=" LF_MAC_FORMAT,
	p, LF_MAC_ARGS(pep->pe_peer_mac), LF_MAC_ARGS(pep->pe_gateway_mac));
  }
}
