#ifndef _fma_myrioe_h_
#define _fma_myrioe_h_

#include <time.h>
#include "fma_myri.h"
#include "fma_myri_packet.h"

/*
 * Information about the Myrinet interfaces on this node
 */
struct fma_myrioe_info {
  struct lf_nic **moe_peers; /* peer table */
  int moe_num_peers;		/* count of peers seen */
  int moe_peers_size;		/* alloced size of peer table */

  struct moe_peer_entry *moe_peer_table;
  int moe_peer_table_len;

  struct lf_event *moe_start_delay_timer;

  struct lf_event *moe_broadcast_task;
  int moe_broadcast_count;

  /* mapper uses this as a timer for sending verify packets */
  struct lf_event *moe_verify_task;

  /* non-mapper uses as a watchdog timer for receiving verify packets */
  struct lf_event *moe_mapper_watchdog;

  /* Info on current MoE mapping process */
  int moe_mapping_in_progress;
  unsigned int moe_map_version;
  int moe_last_rank;			/* used to compute map holdoff */

  /* for mappers specifically */
  int moe_map_reply_flag;

  /* list of mapper scout ACK records */
  struct moe_scout_ack_record *scout_acks;
};

struct moe_scout_ack_record {
  lf_mac_addr_t sa_scouter_mac;
  int sa_scouter_port;
  int sa_rx_nic_id;
  int sa_rx_port;

  int sa_map_version;
  int sa_map_ack_pending;		/* true if scout reply ack pending */

  struct moe_scout_ack_record *next;
};


#define MOE_MAP_START_DELAY 8000
#define MOE_MAP_BROADCAST_COUNT 70
#define MOE_MAP_BROADCAST_INTERVAL 500

#define MOE_VERIFY_BROADCAST_INTERVAL (1000*8)

/* Add 100ms per rank so that the next expected mapper will be the first
 * guy to notice the old mapper is gone
 */
#define MOE_MAPPER_WATCHDOG_TIMEOUT(R) \
  ((MOE_VERIFY_BROADCAST_INTERVAL*6+1) + (R)*100)

/*
 * packets
 */
#define ETHER_TYPE_MX   0x86DF
#define MYRI_TYPE_MX    0x0022
#define MYRI_TYPE_ETHER 0x0009

/* 32 Bytes */
struct mome_pkt_hdr {
  union {

    /* MoE header */
    struct {
      uint8_t dest_mac[6];
      uint8_t src_mac[6];
      uint16_t proto_16;
      uint16_t sender_peer_index_16;
    } e;

    /* MoM header */
    struct {
      uint16_t proto_16;	/* == MYRI_TYPE_MX */
      uint16_t sender_peer_index_16;
      uint16_t dest_mac_low_16;
      uint16_t src_mac_high_16;
      uint32_t src_mac_low_32;
      uint32_t dest_mac_high_32;
    } m;
  } h;

  uint8_t pkt_type;
  uint8_t chargap[3];
  uint32_t gap[3];              /* pad to 32 bytes */
};

/*
 * EoM header, which we will see when receiving MoE broadcast on MoM through
 * a gateway
 */
struct myri_ether_pkt {
  uint16_t type;		/* == MYRI_TYPE_ETHER */
  uint16_t subtype;
  uint16_t payload_len;
  uint16_t seqnum;
  uint8_t sender_mac[6];
};


/*
 * generic MoE packet
 */
struct moe_pkt {
  struct mome_pkt_hdr hdr;	/* 32 bytes */
  uint16_t moe_pkt_type_16;
  uint16_t moe_pkt_subtype_16;
  uint8_t pad[28];
};

/*
 * An MoE broadcast scout packet
 * FMA_SUBTYPE_MOE_BC_SCOUT
 */
struct moe_bc_scout_pkt {
  struct mome_pkt_hdr hdr;	/* 32 bytes */

  uint16_t moe_pkt_type_16;
  uint16_t moe_pkt_subtype_16;
  uint32_t moe_map_version_32;	
  uint8_t moe_scouter_mac_addr[6];
  uint8_t moe_scouter_max_mac_addr[6];
  uint8_t moe_gateway_mac_addr[6];
  uint8_t moe_level_8;
  uint8_t moe_fw_type;
  uint8_t moe_num_ports;
  uint8_t moe_tx_port;
  uint8_t moe_scout_type;
  uint8_t pad[1];             /* pad to 64 bytes */
};

/*
 * An MoE broadcast scout packet reply
 * FMA_SUBTYPE_MOE_SCOUT_RESP
 */
struct moe_bc_scout_reply_pkt {
  struct mome_pkt_hdr hdr;	/* 32 bytes */

  uint16_t moe_pkt_type_16;
  uint16_t moe_pkt_subtype_16;
  uint32_t moe_map_version_32;
  uint8_t moe_scouter_mac_addr[6];
  uint8_t moe_scoutee_mac_addr[6];
  uint8_t moe_scoutee_max_mac_addr[6];
  uint8_t moe_gateway_mac_addr[6];
  uint8_t moe_scoutee_level_8;
  uint8_t moe_fw_type;
  uint8_t moe_num_ports;
  uint8_t moe_rx_port;		/* port scout rcvd on */
};
  
/*
 * An ACK to an MoE broadcast scout packet reply
 * FMA_SUBTYPE_MOE_ACK
 */
struct moe_ack_pkt {
  struct mome_pkt_hdr hdr;	/* 32 bytes */

  uint16_t moe_pkt_type_16;
  uint16_t moe_pkt_subtype_16;
  uint32_t moe_map_version_32;
  lf_mac_addr_t moe_scouter_mac_addr;
  uint8_t moe_tx_port;
  uint8_t pad[17];		/* pad to 64 bytes */
};

/*
 * A verify packet which is broadcast by the mapper
 * once everything is all settled down
 * FMA_SUBTYPE_MOE_BC_VERIFY
 */
struct moe_bc_verify_pkt {
  struct mome_pkt_hdr hdr;	/* 32 bytes */

  uint16_t moe_pkt_type_16;
  uint16_t moe_pkt_subtype_16;
  uint32_t moe_map_version_32;	
  uint8_t moe_verifier_mac_addr[6];
  uint8_t moe_verifier_max_mac_addr[6];
  uint8_t moe_verifier_level_8;
  uint8_t pad[13];
};
  
/*
 * An request to remap the fabric
 * FMA_SUBTYPE_MOE_REMAP_REQUEST
 */
struct moe_remap_request_pkt {
  struct mome_pkt_hdr hdr;	/* 32 bytes */

  uint16_t moe_pkt_type_16;
  uint16_t moe_pkt_subtype_16;
  lf_mac_addr_t moe_requester_mac_addr;

  /* embedded struct for generic routine to use */
  struct fma_map_request fma_remap_request;
};

/*
 * MoE peer list - this packet is followed by
 * moe_num_peers_32 * struct moe_peer_entry
 *
 * Since this arrives via tunnel, it has no mome_pkt_hdr
 *
 * FMA_SUBTYPE_MOE_PEER_LIST
 */
struct moe_peer_list_pkt {
  struct lf_myri_packet_hdr hdr;
  uint32_t moe_num_peers_32;
  uint32_t moe_map_version_32;
};

struct moe_peer_entry {
  uint8_t pe_peer_mac[6];
  uint8_t pe_gateway_mac[6];
  uint8_t pe_num_ports;
  uint8_t pe_fw_type;
  uint8_t pe_level_8;
  uint8_t pad[1];
};

/*
 * Myri-over-ethernet routine prototypes
 */
int moe_raw_send(struct fma_nic_info *nip, int port,
  lf_mac_addr_t dest_mac_addr, void *vpkt, int pkt_len,
  void (*callback)(void *), void *arg);
int fma_init_nic_moe(int nic_id, int nic_handle);
void moe_start_mapping(void);
void moe_stop_mapping(void);
int mome_raw_send(struct fma_nic_info *nip, int port,
  unsigned char *route, int rlen, lf_mac_addr_t dest_mac,
  void *vpkt, int pkt_len, void (*callback)(void *), void *arg);
void moe_check_mom_packet(struct fma_nic_info *nip, int port,
  struct mome_pkt_hdr *pkt, int length);
void mome_handle_pkt(struct fma_nic_info *nip, int port,
  lf_mac_addr_t gateway_mac, struct moe_pkt *pkt, int pkt_len);
int fma_moeom_generic_raw_send(struct fma_generic_send_info *gsip,
    void *p, int len, void (*callback)(void *), void *context);
int fma_moe_generic_raw_send(struct fma_generic_send_info *gsip,
    void *p, int len, void (*callback)(void *), void *context);
void moe_got_peer_list(struct moe_peer_list_pkt *pkt, lf_mac_addr_t sender_mac);
void moe_distribute_peer_table(void);
void moe_build_peer_table(void);
int mome_broadcast_through_gateways(struct fma_nic_info *lcl_nip,
  int lcl_port, void *vpkt, int pkt_len,
  unsigned char *gw_mac_addr_location);

/*
In the E->M direction:

    MXoE unicasts to NICs on the local Myrinet are header swapped.

    All other MXoE packets become MXoEoM.

        Broadcasts are forwarded as MXoEoM to all NICs and GWs.
        (They have to be EoM to be forwarded to gateways,
         and I was lazy and left them that way to NICs, to.
         What luck!)
       
        Unicasts to unknown destinations are forwarded as MXoEoM
        to all gateways.  (They have to be EoM to go to
        gateways.)  These are not forwarded to NICs.

    In the M->E direction:

    All MXoM packets are swapped to MXoE and forwarded onto the
    Ethernet. 
*/

#endif /* _fma_myrioe_h_ */
