/*
 * Manage the list of enclosures in the database
 */
#include <unistd.h>
#include <stdio.h>

#include "db.h"
#include "libfma.h"
#include "lf_dflt_error.h"
#include "lf_fabric.h"
#include "lf_fabric_db.h"

void
usage(
  int code)
{
  fprintf(stderr, "fm_dbdelete is used to delete entries from the\n");
  fprintf(stderr, "fabric database.  This command should not be used to modify the database\n");
  fprintf(stderr, "while fm_server is running - it is intended as a setup command\n");
  fprintf(stderr, "to be used before fm_server is started.\n\n");
  fprintf(stderr, "Usage: fm_dbdelete\n");
  fprintf(stderr, "  -h <hostname> - delete host <hostname>\n");
  fprintf(stderr, "  -n <hostname> <nic_id> - delete NIC <nic_id> from host <hostname>\n");
  fprintf(stderr, "  -s <switch> - delete switch <switch>\n");
  fprintf(stderr, "  -l <switch> <slot> - delete the linecard in slot <slot> of <switch>\n");
  fprintf(stderr, "  -H - print this help message\n");
  fprintf(stderr, "  -R <fms_run> - specify FMS_RUN directory\n");
  fprintf(stderr, "  -N <fms_db_name> - specify database name\n");
  fprintf(stderr, "  -V - print version\n");
  exit(code);
}

void
remove_nic(
  struct lf_fabric_db *fdp,
  char *hostname,
  char *slot_name)
{
  int rc;
  int slot;

  slot = atoi(slot_name);

  rc = lf_remove_link_from_db(fdp, hostname, slot, -1, -1);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s:%d's links from DB\n", hostname, slot);
    exit(1);
  }
  rc = lf_remove_nic_from_db(fdp, hostname, slot);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s, NIC %d from DB\n",
	    hostname, slot);
    exit(1);
  }

  printf("Removed NIC %s:%d from DB.\n", hostname, slot);
}

void
remove_linecard(
  struct lf_fabric_db *fdp,
  char *enc_name,
  char *slot_name)
{
  int rc;
  int slot;

  slot = atoi(slot_name);

  rc = lf_remove_link_from_db(fdp, enc_name, slot, -1, -1);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s's links from DB\n", enc_name);
    exit(1);
  }
  rc = lf_remove_linecard_from_db(fdp, enc_name, slot);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s, slot %d from DB\n",
	    enc_name, slot);
    exit(1);
  }

  printf("Removed linecard %s:%d from DB.\n", enc_name, slot);
}

void
remove_host(
  struct lf_fabric_db *fdp,
  char *hostname)
{
  int rc;

  rc = lf_remove_link_from_db(fdp, hostname, -1, -1, -1);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s's links from DB\n", hostname);
    exit(1);
  }
  rc = lf_remove_nic_from_db(fdp, hostname, -1);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s's NICs from DB\n",
	    hostname);
    exit(1);
  }
  rc = lf_remove_host_from_db(fdp, hostname);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s from DB\n", hostname);
    exit(1);
  }

  printf("Removed host %s from DB.\n", hostname);
}

void
remove_switch(
  struct lf_fabric_db *fdp,
  char *enc_name)
{
  int rc;

  rc = lf_remove_link_from_db(fdp, enc_name, -1, -1, -1);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s's links from DB\n", enc_name);
    exit(1);
  }
  rc = lf_remove_linecard_from_db(fdp, enc_name, -1);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s's lincards from DB\n",
	    enc_name);
    exit(1);
  }
  rc = lf_remove_enclosure_from_db(fdp, enc_name);
  if (rc != 0) {
    fprintf(stderr, "Error removing %s from DB\n", enc_name);
    exit(1);
  }

  printf("Removed switch %s from DB.\n", enc_name);
}

int
main(
  int argc,
  char **argv)
{
  int i;
  int rc;
  char *db_name;
  char *fms_run;
  db_database_ptr_t dbp;
  struct lf_fabric *fp;
  struct lf_fabric_db *fdp;

  lf_init();

  /* check environment variables */
  fms_run = getenv(LF_ENV_FMS_RUN);
  if (fms_run == NULL || *fms_run == '\0') {
    fms_run = LF_DFLT_FMS_RUN;
  }
  db_name = getenv(LF_ENV_DB_NAME);
  if (db_name == NULL || *db_name == '\0') {
    db_name = LF_DFLT_DB_NAME;
  }


  /* command line args */
  for (i=1; i<argc; ++i) {
    if (argv[i][0] == '-') {
      switch (argv[i][1]) {
      case 'H':
	usage(0);
	break;
      case 'R':
	if (++i >= argc) usage(1);
	fms_run = argv[i];
	break;
      case 'N':
	if (++i >= argc) usage(1);
	db_name = argv[i];
	break;
      case 'V':
	printf("FMS version is %s\n", Lf_version);
	exit(0);
	break;

      case 'n':
      case 'l':
	++i;
	/* FALLSTHROUGH */
      case 'h':
      case 's':
	++i;
	if (i >= argc) usage(1);
	break;
      default:
	usage(1);
	break;
      }
    } else {
      usage(2);
    }
  }

  /* open the database now that we have overrides */
  dbp = db_open_database(fms_run, db_name, 1);
  if (dbp == NULL) {
    LF_ERROR(("Error opening database, %s needs to exist and be writeable.",
	  fms_run));
  }

  fdp = lf_fabric_db_handle(dbp);
  if (fdp == NULL) LF_ERROR(("Error getting fabric db handle"));

  fp = lf_load_fabric(fdp);
  if (fp == NULL) {
    fprintf(stderr, "Error opening database at %s/%s\n", fms_run, db_name);
    exit(1);
  }


  /*
   * Now, re-scan the arguments and take action on the opened database
   */
  /* command line args */
  for (i=1; i<argc; ++i) {
    if (argv[i][0] == '-') {
      switch (argv[i][1]) {
      case 's':
	if (++i >= argc) usage(1);
	remove_switch(fdp, argv[i]);
	break;

      case 'l':
	if (++i >= argc-1) usage(1);
	remove_linecard(fdp, argv[i], argv[i+1]);
	++i;
	break;

      case 'h':
	if (++i >= argc) usage(1);
	remove_host(fdp, argv[i]);
	break;

      case 'n':
	if (++i >= argc-1) usage(1);
	remove_nic(fdp, argv[i], argv[i+1]);
	++i;
	break;
      }
    }
  }

  /* flush the database to disk */
  rc = lf_flush_fabric_db(fdp);
  if (rc != 0) {
    fprintf(stderr, "Error writing DB back to disk\n");
    exit(1);
  }
  exit(0);

 except:
  exit(1);
}
